"""
Using AWS Version 4 signing method to make a request to s3. There are multiple
ways to sign requests to AWS, we're using the Authorization Header method.
http://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-auth-using-authorization-header.html

The basic algorithm is:
    1. Create a "string to sign", a concatenation of select request elements
    2. Create a "signing key", a key created from hashing the secret key and
       date, region, and service (e.g. S3) elements.
    3. Use signing key to sign string to sign, include as Authorization header
       on request.

"""
import hashlib
import hmac
import six
if six.PY3:
    from urllib.parse import quote  # pragma: no cover
    from urllib.parse import quote_plus  # pragma: no cover
else:
    from urllib import quote  # pragma: no cover
    from urllib import quote_plus  # pragma: no cover


ISO8601_FMT = "%Y%m%dT%H%M%SZ"
DATE_FMT = "%Y%m%d"


def compute_hashed_payload(payload):
    return hashlib.sha256(payload).hexdigest()


def sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def create_canonical_query_string(query_params):
    """Canonical query params for S3 requests are all url-encoded and sorted
    by key name. E.g. {"mimeType": "application/json", "limit": 20} will yield
    a query_string of "limit=20&mimeType=application%2Fjson"

    :param query_params: dict of query params
    :type query_params: dict with str keys
    """
    string_params = []
    for key in sorted(query_params.keys()):
        if not isinstance(key, six.string_types):
            raise ValueError("Query param keys must be strings")
        string_params.append(
            quote_plus(key) + "=" + quote_plus(query_params[key])
        )
    return "&".join(string_params)


def create_canonical_headers(headers, bucket, region):
    """Canonical headers is a list of certain request headers and values,
    separated by newlines, adhering to some AWS-specific rules. Signed headers
    is just a sorted list of all included canonical headers.

    Rules:
    - header names must be lowercase
    - sorted alphabetically by name
    - trailing whitespace in values is trimmed
    - "host" header is always required. In our case, this will always be
      "s3.amazonaws.com"
    - "x-amz-content-sha256" (hash of payload) is always required. If there is
      no payload, you must provide the hash of the empty string.
    - if "Content-Type" header is present, it must be included
    - all "x-amz-" prefixed headers must be included
    - any additional headers *can* be included. We will include them, as AWS
      docs claim this adds increased security.
    """
    headers = dict(
        (name.lower(), value.strip())
        for name, value in six.iteritems(headers)
    )

    sorted_headers = sorted(six.iteritems(headers))
    signed_headers = ";".join(
        name for (name, value) in sorted_headers
    )
    canonical_headers = "\n".join(
        name + ":" + value for (name, value) in sorted_headers
    )
    canonical_headers += "\n"
    return signed_headers, canonical_headers


def canonicalize_request(
    method,
    path,
    bucket,
    region,
    headers,
    canonical_query_string,
    hashed_payload,
):
    """Creates a the "canonical request" used in signing AWS sigv4 requests.
    See http://docs.aws.amazon.com/AmazonS3/latest/API/sig-v4-header-based-auth.html#canonical-request  # noqa

    HTTPMethod + "\n" +
    CanonicalURI + "\n" +
    CanonicalQueryString + "\n" +
    CanonicalHeaders + "\n" +
    SignedHeaders + "\n" +
    HexEncode(Hash(RequestPayload))

    Note that this is simply a way of organizing and serializing the HTTP request
    for authorization purposes.
    """
    canonical_uri = quote(path)
    signed_headers, canonical_headers = create_canonical_headers(
        headers,
        bucket,
        region,
    )

    elements = [
        method,
        canonical_uri,
        canonical_query_string,
        canonical_headers,
        signed_headers,
        hashed_payload,
    ]
    return "\n".join(elements), signed_headers


def create_string_to_sign(canonical_request, region, dt):
    """The string to sign is a string in a specific format:
    "AWS4-HMAC-SHA256" + "\n" +
    ISO-8601 timestamp + "\n" +
    Scope + "\n" +
    Hex(SHA256Hash(CanonicalRequest))

    Scope is date + "/" + region + "/" + service + "/" + "aws4_request"
    """
    scope_elements = [
        dt.strftime(DATE_FMT),
        region,
        "s3",
        "aws4_request",
    ]
    scope = "/".join(scope_elements)
    string_elements = [
        "AWS4-HMAC-SHA256",
        dt.strftime(ISO8601_FMT),
        scope,
        hashlib.sha256(canonical_request.encode("utf-8")).hexdigest(),
    ]
    string_to_sign = "\n".join(string_elements)
    return string_to_sign, scope


def create_signing_key(now, secret_key, region):
    date_key = sign(("AWS4" + secret_key).encode("utf-8"), now.strftime("%Y%m%d"))
    region_key = sign(date_key, region)
    service_key = sign(region_key, "s3")
    return sign(service_key, "aws4_request")


def format_auth_header(access_key, scope, signed_headers, signature):
    return (
        "AWS4-HMAC-SHA256 Credential={access_key}/{scope}, "
        "SignedHeaders={signed_headers}, Signature={signature}"
    ).format(
        access_key=access_key,
        scope=scope,
        signed_headers=signed_headers,
        signature=signature,
    )


def compute_auth_header(
    headers,
    method,
    dt,
    region,
    bucket,
    path,
    canonical_query_string,
    hashed_payload,
    access_key,
    secret_key,
):
    # Format HTTP request for signing
    canonical_request, signed_headers = canonicalize_request(
        method,
        path,
        bucket,
        region,
        headers,
        canonical_query_string,
        hashed_payload,
    )
    string_to_sign, scope = create_string_to_sign(canonical_request, region, dt)
    # Sign request
    signing_key = create_signing_key(dt, secret_key, region)
    signature = hmac.new(
        signing_key,
        string_to_sign.encode("utf-8"),
        hashlib.sha256,
    ).hexdigest()
    # Format Authorization header
    return format_auth_header(
        access_key,
        scope,
        signed_headers,
        signature,
    )



# -*- coding: utf-8 -*-
import re
from xml.etree import ElementTree


class S3Response(object):
    """An object representing an HTTP response from S3. This is for simple
    HTTP responses that don't need extra parsing. The only two attributes it
    exposes officially are .code (response code) and .body (response body).

    Unofficially, however, the full HTTP response can be accessed using
    ._response. This will be a `fido.Response` object. You can use this to
    access the response's headers.
    """
    def __init__(self, response):
        self._response = response

    @property
    def code(self):
        """HTTP response code"""
        return self._response.code

    @property
    def body(self):
        """HTTP response body"""
        return self._response.body


class ListResponse(S3Response):
    """An object representing parsed XML returned by List Objects v2 endpoint.
    http://docs.aws.amazon.com/AmazonS3/latest/API/v2-RESTBucketGET.html

    Example usage:
    client = S3Client(...)
    response = S3Client.list().result()
    print(response.keys)  # list of strings
    """

    # S3's XML responses are namespaced, such that the XML tags always begin
    # with, e.g. "{http://s3.amazonaws.com/doc/2006-03-01/}". Instead of
    # hardcoding these tags, we'll use a regex to extract them, future-proofing
    # against them changing the tag.
    NS_PATTERN = re.compile(r"\{http.*?\}")

    def __init__(self, response):
        super(ListResponse, self).__init__(response)

        self._root = ElementTree.fromstring(response.body)
        self._namespace = ListResponse._get_namespace(self._root)

        self._is_truncated = ListResponse._get_is_truncated(
            self._root,
            self._namespace,
        )
        self._continuation_token = None
        if self._is_truncated:
            self._continuation_token = ListResponse._get_continuation_token(
                self._root,
                self._namespace,
            )

        self._keys = None
        self._common_prefixes = None

    @property
    def is_truncated(self):
        """If there are more keys/common_prefixes than are returned as part of
        this result. If true, more HTTP calls will have to be made using
        continuation_token to get the bucket's entire contents. See the docstring
        for continuation_token for more info.
        """
        return self._is_truncated

    @property
    def continuation_token(self):
        """A special token that only AWS understands, used to paginate results.

        Example:
            r = client.list(limit=10).result()
            while r.is_truncated:
                print(r.keys)
                r = client.list(limit=10, continuation_token=r.continuation_token).result()  # noqa
        """
        return self._continuation_token

    @property
    def keys(self):
        """Object paths returned by this list operation. Note that if
        a delimiter is specified, this property will *not* include keys rolled
        up under common_prefixes. For more information see the common_prefixes
        docstring.
        """
        if self._keys is None:
            self._keys = [
                element.text for element in
                self._root.findall(
                    "{ns}Contents/{ns}Key".format(ns=self._namespace)
                )
            ]
        return self._keys

    @property
    def common_prefixes(self):
        """List of common prefixes. This attribute will only have entries if
        a delimiter is specified. In this case, keys that share a common prefix
        will be left out of .keys. See the AWS documentation for more information.

        As a concrete example, imagine your bucket contains the following objects:
        a/b/1
        a/b/2
        a/3
        a/4

        Example with delimiter::
            r = client.list(prefix="a/", delimiter="/").result()
            print(r.keys)  # prints ["a/3", "a/4"]
            print(r.common_prefixes)  # prints ["a/b/"]

        Example without delimiter::
            r = client.list(prefix="path/").result()
            print(r.keys)  # prints ["a/b/1", "a/b/2", "a/3", "a/4"]
            print(r.common_prefixes)  # prints []
        """
        if self._common_prefixes is None:
            self._common_prefixes = [
                element.text for element in
                self._root.findall(
                    "{ns}CommonPrefixes/{ns}Prefix".format(ns=self._namespace)
                )
            ]
        return self._common_prefixes

    @classmethod
    def _get_namespace(cls, element):
        match = cls.NS_PATTERN.match(element.tag)
        return match.group()

    @staticmethod
    def _get_is_truncated(root, namespace):
        # The relevant XML tag looks like <IsTruncated>false</IsTruncated>
        return root.find(
            "{ns}IsTruncated".format(ns=namespace)
        ).text == "true"

    @staticmethod
    def _get_continuation_token(root, namespace):
        return root.find(
            "{ns}NextContinuationToken".format(ns=namespace)
        ).text



# -*- coding: utf-8 -*-
import datetime
import logging

import fido

from twisted_s3 import auth
from twisted_s3 import response
from twisted_s3.future import Future


US_EAST_TEMPLATE = "{bucket}.s3.amazonaws.com"
HOST_TEMPLATE = "{bucket}.s3-{region}.amazonaws.com"

# Turn down default logging in twisted. Ideally this would be fixed upstream,
# but that ticket has been open for 7 years:
# http://twistedmatrix.com/trac/ticket/4021
logging.getLogger('twisted').setLevel(logging.CRITICAL)


class S3Client(object):

    def __init__(
        self, access_key, secret_key,
        region=None, bucket=None,
    ):
        """An asynchronous s3 client based on the fido HTTP utility,
        which is based on twisted/crochet.

        :param access_key: AWS access key ID
        :type access_key: str
        :param secret_key: AWS secret
        :type secret_key: str
        :param region: AWS region
        :type region: str, default None
        :param bucket: AWS bucket
        :type bucket: str, default None
        """
        self.access_key = access_key
        self.secret_key = secret_key
        self.region = region
        self.bucket = bucket

    def _make_request(
        self, method, path, headers, query_params, payload,
        region=None, bucket=None,
    ):
        """Authenticate and actually make the HTTP request to S3."""
        headers = headers or {}
        query_params = query_params or {}
        region = region or self.region
        bucket = bucket or self.bucket

        if not region or not bucket:
            raise ValueError(
                "Region and bucket must be either set at the client level "
                "or passed in at call time. "
                "Region={region}; Bucket={bucket}".format(
                    region=region,
                    bucket=bucket,
                ),
            )

        hashed_payload = auth.compute_hashed_payload(payload)

        host_template = US_EAST_TEMPLATE if region == "us-east-1"\
            else HOST_TEMPLATE
        host = host_template.format(bucket=bucket, region=region)

        query_string = auth.create_canonical_query_string(query_params)
        if not path.startswith("/"):
            path = "/" + path

        now = datetime.datetime.utcnow()

        # Add headers necessary for auth computation
        headers["host"] = host
        headers["x-amz-content-sha256"] = hashed_payload
        headers["x-amz-date"] = now.strftime(auth.ISO8601_FMT)
        headers["Authorization"] = auth.compute_auth_header(
            headers,
            method,
            now,
            region,
            bucket,
            path,
            query_string,
            hashed_payload,
            self.access_key,
            self.secret_key,
        )

        url = "http://{host}{path}".format(host=host, path=path)
        if query_string:
            url += "?" + query_string
        return fido.fetch(url, method=method, body=payload, headers=headers)

    def get(
        self,
        path,
        headers=None,
        query_params=None,
        region=None,
        bucket=None,
    ):
        """Get an object from S3. See
        http://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectGET.html

        Example usage:
        client = S3Client(...)
        future = S3Client.get("/test/key")
        key = future.result()
        print(key.body)  # prints the object's bytes

        :param path: Object path, e.g. "/logs/2016/file.gz"
        :type path: str
        :param headers: Additional headers to send in S3 HTTP request. Note
            that the headers "x-amz-date" and "Authorization" will be added as
            part of the authentication process.
        :type headers: dict, with str keys
        :param query_params: Query parameters to include in the request. This
            is rarely used, but can be useful for, say, getting a specified
            version of an object.
        :type query_params: dict
        :param region: S3 region in which the object resides. E.g. "us-west-2".
        :type region: str
        :param bucket: Name of S3 bucket
        :type bucket: str

        :rtype: Future, returning S3Response.
        """
        return Future(
            self._make_request(
                method="GET",
                path=path,
                headers=headers,
                query_params=query_params,
                payload=b"",
                region=region,
                bucket=bucket,
            ),
            response_class=response.S3Response,
        )

    @staticmethod
    def _add_query_params_for_list(
        query_params,
        prefix,
        limit,
        delimiter,
        continuation_token,
    ):
        query_params["list-type"] = "2"

        if prefix is not None:
            query_params["prefix"] = prefix
        if continuation_token is not None:
            query_params["continuation-token"] = continuation_token
        if limit is not None:
            query_params["max-keys"] = str(limit)
        if delimiter is not None:
            query_params["delimiter"] = delimiter

    def list(
        self,
        prefix=None,
        limit=None,
        delimiter=None,
        continuation_token=None,
        headers=None,
        query_params=None,
        region=None,
        bucket=None,
    ):
        """Get a list of objects in a bucket, filtered by certain criteria. See
        http://docs.aws.amazon.com/AmazonS3/latest/API/v2-RESTBucketGET.html

        Example usage:
        client = S3Client(...)
        future = S3Client.list("test/", limit=10)
        results = future.result()
        print(results.keys)  # returns a list of strings

        :param prefix: Prefix to use to filter results. This should be *without*
            the leading slash. E.g. "logs/2016/" not "/logs/2016/".
        :type prefix: str
        :param limit: Max # of keys to return. Default is 1000. When delimiter
            is specified, each instances of CommonPrefix counts as one key.
        :type limit: int
        :param delimiter: A character used to group keys. If specified, all
            keys with the same string between the prefix and the first
            occurence of delimiter will be grouped under .common_prefixes
        :type delimiter: str
        :param continuation_token: Used to paginate results when more than max
            keys (default 1000).
        :type continuation_token: str
        :param headers: Additional headers to send in S3 HTTP request. Note
            that the headers "x-amz-date" and "Authorization" will be added as
            part of the authentication process.
        :type headers: dict, with str keys
        :param query_params: Additional query params. These are used to limit
            the result set. See above link for more information.
        :param region: S3 region in which the object resides. E.g. "us-west-2".
        :type region: str
        :param bucket: Name of S3 bucket
        :type bucket: str

        :rtype: Future, returning ListResponse.
        """
        query_params = query_params or {}
        S3Client._add_query_params_for_list(
            query_params,
            prefix,
            limit,
            delimiter,
            continuation_token,
        )

        return Future(
            self._make_request(
                method="GET",
                path="/",
                headers=headers,
                query_params=query_params,
                payload=b"",
                region=region,
                bucket=bucket,
            ),
            response_class=response.ListResponse,
        )

    def put(
        self,
        path,
        payload,
        headers=None,
        region=None,
        bucket=None,
    ):
        """Set S3 object at path to payload. See
        http://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectPUT.html

        Note that S3 REST API doesn't accept query params for PUTs.

        :param path: Object path, e.g. "/logs/2016/file.gz"
        :type path: str
        :param payload: Object payload. I.e. what to put in S3
        :type payload: bytes
        :param headers: Additional headers to send in S3 HTTP request. Note
            that the headers "x-amz-date" and "Authorization" will be added as
            part of the authentication process.
        :type headers: dict, with str keys and str values
        :param region: S3 region in which the object resides. E.g. "us-west-2".
        :type region: str
        :param bucket: Name of S3 bucket
        :type bucket: str

        :rtype: Future, returning S3Response.
        """
        return Future(
            self._make_request(
                method="PUT",
                path=path,
                headers=headers,
                query_params=None,
                payload=payload,
                region=region,
                bucket=bucket,
            ),
            response_class=response.S3Response,
        )



from __future__ import absolute_import, division, print_function

import os

from setuptools import setup

base_dir = os.path.dirname(__file__)

setup(
    name='twisted_s3',
    packages=['twisted_s3'],
    description='Asynchronous HTTP client for interacting with Amazon S3',
    url='https://github.com/Yelp/twisted_s3',
    author='Yelp Performance Team',
    author_email='no-reply+use_github_issues@yelp.com',
    platforms='all',
    py_modules=['twisted_s3'],
    install_requires=[
        'fido >= 2.1.0',
        'six',
    ],
    extras_require={
        ':python_version!="2.6"': ['twisted >= 14.0.0'],
        ':python_version=="2.6"': ['twisted >= 14.0.0, < 15.5'],
    },
    options={
        'bdist_wheel': {
            'universal': 1,
        },
    },
    version='0.2.2',
)



# -*- coding: utf-8 -*-


class S3ResponseError(Exception):
    """Class for all HTTP errors received from S3. See
    http://docs.aws.amazon.com/AmazonS3/latest/API/ErrorResponses.html for all
    potential errors.
    """
    pass


class Future(object):
    """Wraps a crochet.EventualResult object to provide a simple interface to
    HTTP results.

    Usage:
    future = S3Client.get(...)
    response = future.result(2)
    """
    def __init__(self, eventual, response_class):
        self.eventual = eventual
        self.response_class = response_class

    def result(self, timeout=None):
        """Block for `timeout` seconds on the results of the S3 HTTP call.

        :param timeout: time to block waiting for results of the HTTP call.
        :type timeout: float or int
        :raises crochet.TimeoutError: if the result is not ready before
        `timeout` is hit.
        :raises S3ResponseError: if the HTTP call is not successful, e.g. a
        403 response is received.
        """
        result = self.eventual.wait(timeout)
        if result.code >= 300:
            raise S3ResponseError(
                "S3 returned status code: {code} with body: {body}".format(
                    code=result.code,
                    body=result.body,
                ),
            )
        else:
            return self.response_class(result)
